#include "instance.h"

_instructor::_instructor(const _instructor &b) {
	name = b.name;
	id = b.id;
	semester = b.semester;
	lb = b.lb;
	ub = b.ub;
	pref_list = b.pref_list;
}

_course::_course(const _course &b) {
	name = b.name;
	id = b.id;
	semester = b.semester;
	n_instr = b.n_instr;
	pref_list = b.pref_list;
}

std::ostream& operator<<(std::ostream& os, const _instructor& ins) {
	(os) << "(" << ins.id << ") " << ins.name << ": ";
	(os) << ins.semester << " " << ins.lb << " " << ins.ub << "; ";
	vector<vector<string> >::const_iterator pt;
	std::ostream_iterator < string > out_it(os, ",");
	for (pt = ins.pref_list.begin(); pt != ins.pref_list.end(); pt++) {
		(os) << "[";
		std::copy((*pt).begin(), (*pt).end(), out_it);
		(os) << "\b";
		(os) << "]";
	}
	return os;
}

std::ostream& operator<<(std::ostream& os, const _course& crs) {
	(os) << "(" << crs.id << ") " << crs.name << ": ";
	(os) << crs.semester << " " << crs.n_instr << " ";
	vector<vector<string> >::const_iterator pt;
	std::ostream_iterator < string > out_it(os, ",");
	for (pt = crs.pref_list.begin(); pt != crs.pref_list.end(); pt++) {
		(os) << "[";
		std::copy((*pt).begin(), (*pt).end(), out_it);
		(os) << "\b";
		(os) << "]";
	}
	return os;
}

Instance::Instance(string dirname) {
	name = dirname;
	readCourses();
	readInstructors();
	//printData();
	crsRanking = prepareRankings(_ni, _nc, instructors, this, &Instance::crsID);
	instrRanking = prepareRankings(_nc, _ni, courses, this, &Instance::instrID);
	//printRankings();
	nc=_nc;
	ni=_ni;
}

Instance::~Instance() {
	if (crsRanking != NULL) {
		for (int i = 0; i < _ni; i++)
			delete[] crsRanking[i];
		delete[] crsRanking;
	}
	if (instrRanking != NULL) {
		for (int c = 0; c < _nc; c++)
			delete[] instrRanking[c];
		delete[] instrRanking;
	}
}

const int Instance::crsID(string name) const {
	std::map<string, int>::const_iterator search = crs_name_to_id.find(name);
	if (search != crs_name_to_id.end()) {
		return search->second;
	} else {
		std::cerr << "Course " << name << "not found" << std::endl;
		exit(0);
	};
}

const int Instance::instrID(string name) const {
	std::map<string, int>::const_iterator search = instr_name_to_id.find(name);
	if (search != instr_name_to_id.end()) {
		return search->second;
	} else {
		std::cerr << "Instructor " << name << "not found" << std::endl;
		exit(0);
	};
}

void Instance::parse_preferences(string pref,
		vector<vector<string> > &pref_list) {
	string buf;
	string::iterator pc = pref.begin();
	while (pc != pref.end()) {
		if (*pc == '(') {
			vector < string > ties;
			pc++;
			while (*pc != ')') {
				if (*pc == ',') {
					ties.push_back(buf);
					buf.clear();
				} else
					buf.push_back(*pc);
				//cout<<*pc<<"-";
				pc++;
			}
			ties.push_back(buf);
			pref_list.push_back(ties);
			ties.clear();
			buf.clear();
			pc++;
		}
		if (*pc == ',') {
			if (buf.size() > 0) {
				vector < string > t;
				t.push_back(buf);
				pref_list.push_back(t);
				buf.clear();
			}
		} else if (pc == pref.end())
			break;
		else
			buf.push_back(*pc);
		pc++;
	}
	if (buf.size() > 0) {
		vector < string > t;
		t.push_back(buf);
		pref_list.push_back(t);
		buf.clear();
	}
}

void Instance::readInstructors() {
	std::ifstream in;
	string fileinstance;

	fileinstance = name + "/instructors.csv";
	in.open(fileinstance.c_str());
	if (!in) {
		cerr << "Couldn't open the file " << fileinstance << endl;
		exit(1);
	}

	cout << "Read " << fileinstance << endl;

	instructors.clear();
	int id = 0;
	int linenum = 0;
	string line;
	Csv csv(in, ";");
	while (csv.getline(line) != 0) {
		linenum++;
		_instructor tmp_i;
		tmp_i.name = csv.getfield(0);
		instr_name_to_id.insert(pair<string,int> (tmp_i.name, id));
		tmp_i.id = id++;
		tmp_i.semester = atoi(csv.getfield(1).c_str());
		tmp_i.lb = atoi(csv.getfield(2).c_str());
		tmp_i.ub = atoi(csv.getfield(3).c_str());
		parse_preferences(csv.getfield(4), tmp_i.pref_list);
		instructors.push_back(tmp_i);
	}
	_ni = id;
	cout << "Done. Line read (linenum): " << linenum << endl;
}

void Instance::readCourses() {
	ifstream in;
	string fileinstance;

	fileinstance = name + "/courses.csv";
	in.open(fileinstance.c_str());
	if (!in) {
		cerr << "Couldn't open the file " << fileinstance << endl;
		exit(1);
	}

	cout << "Read " << fileinstance << endl;

	courses.clear();
	int linenum = 0;
	int id = 0;
	string line;
	Csv csv(in, ";");
	while (csv.getline(line) != 0) {
		linenum++;
		_course tmp_c;
		tmp_c.name = csv.getfield(0);
		crs_name_to_id.insert(pair<string, int> (tmp_c.name, id));
		tmp_c.id = id++;
		tmp_c.semester = atoi(csv.getfield(1).c_str());
		tmp_c.n_instr = atoi(csv.getfield(2).c_str());
		parse_preferences(csv.getfield(3), tmp_c.pref_list);
		courses.push_back(tmp_c);
	}
	_nc = id;

	cout << "Done. Line read (linenum): " << linenum << endl;
}

template<class X> int ** prepareRankings(int subjects, int objects,
		vector<X> &entities, const Instance *ptr_class, ptr_ID_type ID) {
	int **matrix = new int *[subjects];
	for (int i = 0; i < subjects; i++) {
		matrix[i] = new int[objects];
	}
	vector<vector<string> >::const_iterator pt;
	for (int i = 0; i < subjects; i++) {
		int pos = 1, value = 1;
		vector<bool> checked(objects, false);
		for (pt = entities[i].pref_list.begin();
				pt != entities[i].pref_list.end(); pt++) {
			value = pos + (*pt).size() / 2;
			pos += (*pt).size();
			vector<string>::const_iterator ps;
			for (ps = (*pt).begin(); ps != (*pt).end(); ps++) {
				matrix[i][(*ptr_class.*ID)(*ps)] = value;
				checked[(*ptr_class.*ID)(*ps)] = true;
			}
		}
		// insert all remaining
		if (pos <= objects) {
			value = (objects - pos) / 2;
			for (int c = 0; c < objects; c++)
				if (!checked[c])
					matrix[i][c] = value;
		}
	}
	return matrix;
}

void Instance::printData() {
	std::ostream_iterator<_course> out_crs(cout, "\n");
	std::ostream_iterator<_instructor> out_instr(cout, "\n");
	copy(courses.begin(), courses.end(), out_crs);
	copy(instructors.begin(), instructors.end(), out_instr);
}

const void Instance::printRankings() const {
	cout << "int crsRanking[] = {\n";
	for (int i = 0; i < _ni; i++) {
		cout<<instructors[fromVirtual2RealInstr(i)].name<<":\t";
		for (int c = 0; c < _nc; c++) {
			cout << crsRanking[i][c] << ", ";
		}
		cout << endl;
	}
	cout << "}" << endl;
	cout << "int instrRanking[] = {\n";
	for (int c = 0; c < _nc; c++) {
		cout<<courses[fromVirtual2RealInstr(c)].name<<":\t";
		for (int i = 0; i < _ni; i++) {
			cout << instrRanking[c][i] << ", ";
		}
		cout << endl;
	}
	cout << "}" << endl;
}

const std::vector<int> Instance::instr2IDs(const string name) const {
	vector<int> tmp;
	RangeType ret;
    ret = instr_name_to_id.equal_range(name);
    //std::cout << ch << " =>";
    for (DictionaryType::const_iterator it=ret.first; it!=ret.second; ++it)
      tmp.push_back( it->second );
    return tmp;
}

const std::vector<int> Instance::crs2IDs(const string name) const {
	vector<int> tmp;
	RangeType ret;
    ret = crs_name_to_id.equal_range(name);
    //std::cout << ch << " =>";
    for (DictionaryType::const_iterator it=ret.first; it!=ret.second; ++it)
      tmp.push_back( it->second );
    return tmp;
}


FlatInstance::FlatInstance(string dirname) :
		Instance(dirname) {

	instr_name_to_id.clear();
	ni = 0;
	_fromVirtual2RealInstr.clear();
	for (int j = 0; j < _ni; j++) {
		for (int u = 0; u < instructors[j].ub; u++) {
			_fromVirtual2RealInstr.push_back(j);
			instr_name_to_id.insert(std::pair<string,int> (instructors[j].name,ni) );
			ni++;
		}
	}

	crs_name_to_id.clear();
	nc = 0;
	_fromVirtual2RealCrs.clear();
	for (int j = 0; j < _nc; j++) {
		for (int u = 0; u < courses[j].n_instr; u++) {
			_fromVirtual2RealCrs.push_back(j);
			crs_name_to_id.insert(std::pair<string,int> (courses[j].name,nc) );
			nc++;
		}
	}
	cout<<"Flat instance: nc: "<<nc<<" ni: "<<ni<<endl;
}


