/*
 Instructors: Marco
 */
#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <vector>
#include <map>
#include <iterator>
#include <cstring>
#include <algorithm>
#include "csvgetlinec++.h"

using std::string;
using std::vector;
using std::map;
using std::cout;
using std::cerr;
using std::endl;

typedef std::multimap<string,int> DictionaryType;
typedef std::pair<DictionaryType::const_iterator, DictionaryType::const_iterator> RangeType;


class _instructor {
public:
	_instructor() {
	}
	;
	~_instructor() {
	}
	;
	_instructor(const _instructor &);
	string name;
	int id;
	int semester;
	int lb;
	int ub;
	vector<vector<string> > pref_list;
};

std::ostream& operator<<(std::ostream& os, _instructor& i);

class _course {
public:
	_course() {
	}
	;
	~_course() {
	}
	;
	_course(const _course &);
	string name;
	int id;
	int semester;
	int n_instr;
	vector<vector<string> > pref_list;
};

std::ostream& operator<<(std::ostream& os, _course& c);

class Instance;

typedef const int (Instance::*ptr_ID_type)(string) const;
template<class X> int ** prepareRankings(int subjects, int objects,
		vector<X> &entities, const Instance *, ptr_ID_type ID);

class Instance {
public:
	string name;
	vector<_course> courses;
	int nc;
	vector<_instructor> instructors;
	int ni;

	Instance() {
	}
	;
	~Instance();
	Instance(string dirname);
	void printData();
	const void printRankings() const;
	const int crsID(string name) const;
	const int instrID(string name) const;

	virtual const int fromVirtual2RealCrs(const int j) const {
		return j;
	}
	;
	virtual const int fromVirtual2RealInstr(const int j) const {
		return j;
	}
	;

	virtual const int * InstrRanking(const int c) const {
		return instrRanking[c];
	}
	;
	virtual const int InstrRanking(const int c, const int i) const {
		// returns the array of instructor rank values for a course
		return instrRanking[c][i];
	}
	;
	virtual const int * CrsRanking(const int i) const {
		// returns the array of course rank values for an instructor
		return crsRanking[i];
	}
	;
	virtual const int CrsRanking(const int i, const int c) const {
		return crsRanking[i][c];
	}
	;

	const int realNi(void) const {
		return _ni;
	}
	;

	const std::vector<int> instr2IDs(const string name) const;
	const std::vector<int> crs2IDs(const string name) const;

	DictionaryType crs_name_to_id;
	DictionaryType instr_name_to_id;

private:
	void readInstructors();
	void readCourses();
	void parse_preferences(string pref, vector<vector<string> > &pref_list);

protected:
	int ** instrRanking; // [c][i] for each course c the rank of the instructor i
	int ** crsRanking; // [i][c] for each instructor i the rank of the course c

	int _ni;
	int _nc;



};

class FlatInstance: public Instance {
public:
	vector<int> _fromVirtual2RealCrs;
	vector<int> _fromVirtual2RealInstr;

	FlatInstance() {
	}
	;
	~FlatInstance() {
	}
	;
	FlatInstance(string dirname);

	const int fromVirtual2RealCrs(const int j) const {
		return _fromVirtual2RealCrs[j];
	}
	;
	const int fromVirtual2RealInstr(const int j) const {
		return _fromVirtual2RealInstr[j];
	}
	;

	const int * InstrRanking(const int c) const {
		return instrRanking[_fromVirtual2RealCrs[c]];
	}
	;
	const int InstrRanking(const int c, const int i) const {
		return instrRanking[_fromVirtual2RealCrs[c]][_fromVirtual2RealInstr[i]];
	}
	;
	const int * CrsRanking(const int i) const {
		return crsRanking[_fromVirtual2RealInstr[i]];
	}
	;
	const int CrsRanking(const int i, const int c) const {
		return crsRanking[_fromVirtual2RealInstr[i]][_fromVirtual2RealCrs[c]];
	}
	;
};
